#
#  PyTrain: a library for controlling Lionel Legacy engines, trains, switches, and accessories
#
#  Copyright (c) 2024-2025 Dave Swindell <pytraininfo.gmail.com>
#
#  SPDX-License-Identifier: LPGL
#
#

from __future__ import annotations

from threading import Condition, RLock

from guizero import Text
from guizero.base import Widget
from guizero.event import EventData

from ..protocol.command_req import CommandReq
from ..db.sync_state import SyncState
from ..gui.component_state_gui import ComponentStateGui
from ..gui.state_based_gui import StateBasedGui
from ..protocol.constants import CommandScope, PROGRAM_NAME
from ..protocol.tmcc1.tmcc1_constants import TMCC1SyncCommandEnum


class SystemsGui(StateBasedGui):
    def __init__(
        self,
        label: str = None,
        width: int = None,
        height: int = None,
        aggrigator: ComponentStateGui = None,
        scale_by: float = 1.0,
        hold_for: int = 5,
    ) -> None:
        StateBasedGui.__init__(
            self,
            f"{PROGRAM_NAME} Administration",
            label,
            width,
            height,
            aggrigator,
            enabled_bg="red",
            scale_by=scale_by,
        )
        self._hold_for = hold_for

    # noinspection PyTypeChecker
    def on_sync(self) -> None:
        if self._sync_state.is_synchronized:
            if self._sync_watcher:
                self._sync_watcher.shutdown()
                self._sync_watcher = None
            self._synchronized = True

            # get all target states; watch for state changes
            accs = self.get_target_states()  # should be just 1
            for acc in accs:
                self._states[acc.tmcc_id] = acc
            self._max_name_len = 12

            # start GUI
            self.start()

    def get_target_states(self) -> list[SyncState]:
        pds: list[SyncState] = []
        accs = self._state_store.get_all(CommandScope.SYNC)
        for acc in accs:
            pds.append(acc)
        return pds

    def is_active(self, state: SyncState) -> bool:
        return False

    def switch_state(self, pd: SyncState) -> None:
        pass

    def _make_state_button(
        self,
        pd: SyncState,
        row: int,
        col: int,
    ) -> tuple[list[Widget], int, int]:
        self.by_name.hide()
        self.by_number.hide()
        widgets: list[Widget] = []

        # make title label
        ts = int(round(23 * self._scale_by))
        title = Text(
            self.btn_box,
            text=f"Press for {self._hold_for} seconds",
            grid=[col, row, 2, 1],
            size=ts,
            bold=True,
            color="red",
        )
        widgets.append(title)

        # make reload button
        row += 1
        btn, btn_h, btn_y = super()._make_state_button(pd, row, col)
        btn.text = "Reload Base 3 State"
        safety = PushButtonHoldHelper(self, btn, CommandReq(TMCC1SyncCommandEnum.RESYNC), self._hold_for)
        btn.when_left_button_pressed = safety.on_press
        btn.when_left_button_released = safety.on_release
        self.set_button_inactive(btn)
        widgets.append(btn)

        # make restart button
        row += 1
        btn, btn_h, btn_y = super()._make_state_button(pd, row, col)
        btn.text = "Restart All Nodes"
        safety = PushButtonHoldHelper(self, btn, CommandReq(TMCC1SyncCommandEnum.RESTART), self._hold_for)
        btn.when_left_button_pressed = safety.on_press
        btn.when_left_button_released = safety.on_release
        self.set_button_inactive(btn)
        widgets.append(btn)

        # make update button
        row += 1
        btn, btn_h, btn_y = super()._make_state_button(pd, row, col)
        btn.text = f"Update {PROGRAM_NAME} Software"
        safety = PushButtonHoldHelper(self, btn, CommandReq(TMCC1SyncCommandEnum.UPDATE), self._hold_for)
        btn.when_left_button_pressed = safety.on_press
        btn.when_left_button_released = safety.on_release
        self.set_button_inactive(btn)
        widgets.append(btn)

        # make upgrade button
        row += 1
        btn, btn_h, btn_y = super()._make_state_button(pd, row, col)
        btn.text = "Upgrade Raspberry Pi Software"
        safety = PushButtonHoldHelper(self, btn, CommandReq(TMCC1SyncCommandEnum.UPGRADE), self._hold_for)
        btn.when_left_button_pressed = safety.on_press
        btn.when_left_button_released = safety.on_release
        self.set_button_inactive(btn)
        widgets.append(btn)

        # spacer
        row += 1
        spacer = Text(self.btn_box, text=" ", grid=[col, row], size=ts)
        widgets.append(spacer)

        # make reboot button
        row += 1
        btn, btn_h, btn_y = super()._make_state_button(pd, row, col)
        btn.text = "Reboot All Nodes"
        safety = PushButtonHoldHelper(self, btn, CommandReq(TMCC1SyncCommandEnum.REBOOT), self._hold_for)
        btn.when_left_button_pressed = safety.on_press
        btn.when_left_button_released = safety.on_release
        self.set_button_inactive(btn)
        widgets.append(btn)

        # make shutdown button
        row += 1
        btn, btn_h, btn_y = super()._make_state_button(pd, row, col)
        btn.text = "Shutdown All Nodes"
        safety = PushButtonHoldHelper(self, btn, CommandReq(TMCC1SyncCommandEnum.SHUTDOWN), self._hold_for)
        btn.when_left_button_pressed = safety.on_press
        btn.when_left_button_released = safety.on_release
        self.set_button_inactive(btn)
        widgets.append(btn)

        # noinspection PyTypeChecker
        self._state_buttons[pd.tmcc_id] = widgets
        return widgets, btn_h, btn_y


class PushButtonHoldHelper:
    def __init__(
        self,
        gui: SystemsGui,
        button: Widget,
        command: CommandReq,
        hold_for=5,
    ) -> None:
        self._gui = gui
        self._command = command
        self._button = button
        self._hold_for = hold_for * 1000
        self._app = gui.app
        self._tk = gui.app.tk
        self._after_id = None
        self._cv = Condition(RLock())

    def on_press(self, event: EventData) -> None:
        btn = event.widget
        with self._cv:
            # Cancel previously scheduled call if any
            if self._after_id is not None:
                self._tk.after_cancel(self._after_id)
                self._after_id = None
            # Schedule command for required hold time
            self._gui.set_button_active(btn)
            print(f"Scheduling command {self._command} for {self._hold_for}ms")
            self._after_id = self._tk.after(self._hold_for, self.fire)

    def on_release(self, event: EventData) -> None:
        with self._cv:
            btn = event.widget
            self._gui.set_button_inactive(btn)
            if self._after_id is not None:
                print(f"Cancelling command {self._command}")
                self._tk.after_cancel(self._after_id)
                self._after_id = None

    def fire(self):
        with self._cv:
            self._gui.set_button_inactive(self._button)
            self._after_id = None
            self._command.send()



#
#  PyTrain: a library for controlling Lionel Legacy engines, trains, switches, and accessories
#
#  Copyright (c) 2024-2025 Dave Swindell <pytraininfo.gmail.com>
#
#  SPDX-License-Identifier: LPGL
#
#
import os
import platform
import shutil
import subprocess
import sys
import tempfile
from argparse import ArgumentParser
from pathlib import Path

from .make_base import _MakeBase
from .pytrain import DEFAULT_BUTTONS_FILE
from ..utils.argument_parser import PyTrainArgumentParser
from ..utils.path_utils import find_file


class MakeService(_MakeBase):
    def __init__(self, cmd_line: list[str] = None) -> None:
        self._start_service = False
        super().__init__(cmd_line)

    def program(self) -> str:
        return "make_service"

    def function(self) -> str:
        return "service"

    def postprocess_args(self) -> None:
        if self._args.start is True:
            self._start_service = True
        self._buttons_file = self._args.buttons_file

    def config_header(self) -> list[str]:
        from .. import PROGRAM_NAME

        lines = list()
        lines.append(f"\nInstalling {PROGRAM_NAME} as a systemd service with these settings:")
        lines.append(f"  Start service now: {'Yes' if self._start_service is True else 'No'}")
        return lines

    def install(self) -> None:
        path = self.make_shell_script()
        if path:
            self._config["___SHELL_SCRIPT___"] = str(path)
            self.install_service()

    def remove(self) -> None:
        from .. import PROGRAM_NAME

        if self.is_server_present:
            mode = "Server"
        elif self.is_client_present:
            mode = "Client"
        else:
            print(f"\nNo {PROGRAM_NAME} server or client is currently running. Exiting")
            return
        if self.confirm(f"Are you sure you want to deactivate and remove {PROGRAM_NAME} {mode}?"):
            path = Path(self._home, "pytrain_server.bash" if mode == "Server" else "pytrain_client.bash")
            if path.exists():
                print(f"\nRemoving {path}...")
                path.unlink(missing_ok=True)
            print(f"\nDeactivating {PROGRAM_NAME} {mode} service...")
            self.deactivate_service("pytrain_server" if mode == "Server" else "pytrain_client")

    def command_line_parser(self) -> ArgumentParser:
        from .. import PROGRAM_NAME

        parser = ArgumentParser(add_help=False)
        misc_opts = parser.add_argument_group("Service options")
        misc_opts.add_argument(
            "-buttons_file",
            nargs="?",
            default=None,
            const=DEFAULT_BUTTONS_FILE,
            help=f"Button definitions file, loaded when {PROGRAM_NAME} service starts",
        )
        misc_opts.add_argument(
            "-start",
            action="store_true",
            help=f"Start {PROGRAM_NAME} Client/Server now (otherwise, it starts on reboot)",
        )
        return PyTrainArgumentParser(
            prog=self._prog,
            description=f"Launch {PROGRAM_NAME} as a systemd service when your Raspberry Pi is powered on",
            parents=[self._command_line_parser(), parser],
        )

    def make_shell_script(self) -> Path | None:
        template = find_file("pytrain.bash.template", (".", "../", "src"))
        if template is None:
            print("\nUnable to locate shell script template. Exiting")
            return None
        template_data = ""
        with open(template, "r") as f:
            template_data = f.read()
        for key, value in self.config.items():
            template_data = template_data.replace(key, value)
        path = Path(self._home, "pytrain_server.bash" if self.is_server else "pytrain_client.bash")
        # write the shell script file
        if path.exists():
            shutil.copy2(path, path.with_suffix(".bak"))
        with open(path, "w") as f:
            f.write(template_data)
        os.chmod(path, 0o755)
        print(f"\n{path} created")
        return path

    def install_service(self) -> str | None:
        from .. import PROGRAM_NAME

        if platform.system().lower() != "linux":
            print(f"\nPlease run {self._prog} from a Raspberry Pi. Exiting")
            return None
        template = find_file("pytrain.service.template", (".", "../", "src"))
        if template is None:
            print("\nUnable to locate service definition template. Exiting")
            return None
        template_data = ""
        with open(template, "r") as f:
            template_data = f.read()
        for key, value in self.config.items():
            template_data = template_data.replace(key, value)
        tmp = tempfile.NamedTemporaryFile()
        with open(tmp.name, "w") as f:
            f.write(template_data)
        service = "pytrain_server.service" if self.is_server else "pytrain_client.service"
        result = subprocess.run(
            f"sudo cp -f {tmp.name} /etc/systemd/system/{service}".split(),
            capture_output=True,
            text=True,
        )
        if result.returncode != 0:
            print(f"Error creating /etc/systemd/system/{service}: {result.stderr} Exiting")
            return None
        result = subprocess.run(
            f"sudo chmod 644 /etc/systemd/system/{service}".split(),
            capture_output=True,
            text=True,
        )
        if result.returncode != 0:
            print(f"Error changing mode of /etc/systemd/system/{service}: {result.stderr} Exiting")
            return None
        result = subprocess.run(
            "sudo systemctl daemon-reload".split(),
            capture_output=True,
            text=True,
        )
        if result.returncode != 0:
            print(f"Error reloading system daemons: {result.stderr} Exiting")
            return None
        result = subprocess.run(
            f"sudo systemctl enable {service}".split(),
            capture_output=True,
            text=True,
        )
        if result.returncode != 0:
            print(f"Error enabling {PROGRAM_NAME} service: {result.stderr} Exiting")
            return None
        if self._start_service:
            subprocess.run(
                f"sudo systemctl restart {service}".split(),
            )
            print(f"\n{PROGRAM_NAME} service started...")
        return service

    @property
    def is_server_present(self) -> bool:
        return self.is_service_present("pytrain_server")

    @property
    def is_client_present(self) -> bool:
        return self.is_service_present("pytrain_client")


def main(args: list[str] | None = None) -> int:
    if args is None:
        args = sys.argv[1:]
    try:
        MakeService(args)
        return 0
    except Exception as e:
        # Output anything else nicely formatted on stderr and exit code 1
        sys.exit(f"{__file__}: error: {e}\n")



#
#  PyTrain: a library for controlling Lionel Legacy engines, trains, switches, and accessories
#
#  Copyright (c) 2024-2025 Dave Swindell <pytraininfo.gmail.com>
#
#  SPDX-License-Identifier: LPGL
#
#

from __future__ import annotations

from typing import cast

from guizero import PushButton, Slider, Text
from guizero.base import Widget

from ..pdi.amc2_req import Amc2Req
from ..protocol.tmcc1.tmcc1_constants import TMCC1AuxCommandEnum
from ..protocol.command_req import CommandReq
from ..protocol.constants import CommandScope
from ..db.accessory_state import AccessoryState
from ..gui.component_state_gui import ComponentStateGui
from ..gui.state_based_gui import StateBasedGui
from ..pdi.constants import PdiCommand, Amc2Action


class MotorsGui(StateBasedGui):
    def __init__(
        self,
        label: str = None,
        width: int = None,
        height: int = None,
        aggrigator: ComponentStateGui = None,
        scale_by: float = 1.0,
    ) -> None:
        StateBasedGui.__init__(self, "Motors", label, width, height, aggrigator, scale_by=scale_by)
        self._making_buttons = True

    def _post_process_state_buttons(self) -> None:
        self.app.after(500, self.clear_making_buttons)

    def clear_making_buttons(self) -> None:
        self._making_buttons = False

    @property
    def is_making_buttons(self) -> bool:
        return self._making_buttons

    def get_target_states(self) -> list[AccessoryState]:
        pds: list[AccessoryState] = []
        accs = self._state_store.get_all(CommandScope.ACC)
        for acc in accs:
            acc = cast(AccessoryState, acc)
            if acc.is_amc2 and acc.road_name and acc.road_name.lower() != "unused":
                pds.append(acc)
        return pds

    def is_active(self, state: AccessoryState) -> bool:
        return False

    def switch_state(self, pd: AccessoryState) -> None:
        pass

    # noinspection PyTypeChecker
    def update_button(self, tmcc_id: int) -> None:
        with self._cv:
            pd = self._states[tmcc_id]
            widgets = self._state_buttons[tmcc_id]
            if isinstance(widgets, list):
                for widget in widgets:
                    motor = getattr(widget, "motor", None)
                    lamp = getattr(widget, "lamp", None)
                    if isinstance(widget, PushButton) and motor in {1, 2}:
                        if self.is_motor_active(pd, motor):
                            self.set_button_active(widget)
                        else:
                            self.set_button_inactive(widget)
                    if isinstance(widget, PushButton) and lamp in {1, 2, 3, 4}:
                        if self.is_lamp_active(pd, lamp):
                            self.set_button_active(widget)
                        else:
                            self.set_button_inactive(widget)
                    if isinstance(widget, Slider) and motor in {1, 2}:
                        motor_state = pd.get_motor(motor)
                        if widget.value != motor_state.speed if motor_state else 0:
                            widget.value = motor_state.speed if motor_state else 0.0
                        widget.bg = self._enabled_bg if self.is_motor_active(pd, motor) else "lightgrey"
                    if isinstance(widget, Slider) and lamp in {1, 2, 3, 4}:
                        lamp_state = pd.get_lamp(lamp)
                        if widget.value != lamp_state.level if lamp_state else 0:
                            widget.value = lamp_state.level if lamp_state else 0.0
                        widget.bg = self._enabled_bg if self.is_lamp_active(pd, lamp) else "lightgrey"

    @staticmethod
    def is_motor_active(state: AccessoryState, motor: int) -> bool:
        return state.is_motor_on(state.motor2 if motor == 2 else state.motor1)

    @staticmethod
    def is_lamp_active(state: AccessoryState, lamp: int) -> bool:
        lamp_state = state.get_lamp(lamp)
        return lamp_state and lamp_state.level > 0

    def set_motor_state(self, tmcc_id: int, motor: int, speed: int = None) -> None:
        if self._making_buttons:
            return
        with self._cv:
            pd: AccessoryState = self._states[tmcc_id]
            if speed is None:
                CommandReq(TMCC1AuxCommandEnum.NUMERIC, tmcc_id, data=motor).send()
                if self.is_motor_active(pd, motor):
                    CommandReq(TMCC1AuxCommandEnum.AUX2_OPT_ONE, tmcc_id).send()
                else:
                    CommandReq(TMCC1AuxCommandEnum.AUX1_OPT_ONE, tmcc_id).send()
            elif speed != (pd.motor1.speed if motor == 1 else pd.motor2.speed):
                Amc2Req(tmcc_id, PdiCommand.AMC2_SET, Amc2Action.MOTOR, motor=motor - 1, speed=speed).send()
                CommandReq(TMCC1AuxCommandEnum.NUMERIC, tmcc_id, data=motor).send()
                if speed:
                    CommandReq(TMCC1AuxCommandEnum.AUX1_OPT_ONE, tmcc_id).send()
                else:
                    CommandReq(TMCC1AuxCommandEnum.AUX2_OPT_ONE, tmcc_id).send()

    def set_lamp_state(self, tmcc_id: int, lamp: int, level: int = None) -> None:
        if self._making_buttons:
            return
        with self._cv:
            pd: AccessoryState = self._states[tmcc_id]
            lamp_state = pd.get_lamp(lamp)
            if level is None:
                if self.is_lamp_active(pd, lamp):
                    Amc2Req(tmcc_id, PdiCommand.AMC2_SET, Amc2Action.LAMP, lamp=lamp - 1, level=0).send()
                else:
                    Amc2Req(tmcc_id, PdiCommand.AMC2_SET, Amc2Action.LAMP, lamp=lamp - 1, level=100).send()
                CommandReq(TMCC1AuxCommandEnum.NUMERIC, tmcc_id, data=lamp + 2).send()
            elif level != lamp_state.level:
                Amc2Req(tmcc_id, PdiCommand.AMC2_SET, Amc2Action.LAMP, lamp=lamp - 1, level=level).send()
                CommandReq(TMCC1AuxCommandEnum.NUMERIC, tmcc_id, data=lamp + 2).send()

    def _make_state_button(
        self,
        pd: AccessoryState,
        row: int,
        col: int,
    ) -> tuple[list[Widget], int, int]:
        self._making_buttons = True
        ts = int(round(23 * self._scale_by))
        widgets: list[Widget] = []
        # make title label
        title = Text(self.btn_box, text=f"#{pd.tmcc_id} {pd.road_name}", grid=[col, row, 2, 1], size=ts, bold=True)
        widgets.append(title)

        # make motor 1 on/off button
        row += 1
        m1_pwr, btn_h, btn_y = super()._make_state_button(pd, row, col)
        m1_pwr.text = "Motor #1"
        m1_pwr.motor = 1
        if pd.motor1.state:
            self.set_button_active(m1_pwr)
        m1_pwr.update_command(self.set_motor_state, args=[pd.tmcc_id, 1])
        widgets.append(m1_pwr)

        # motor 1 control
        slider_height = int(round(btn_h * 0.9))
        m1_ctl = Slider(
            self.btn_box,
            grid=[col, row + 1],
            height=slider_height,
            width=self.pd_button_width,
            step=5,
        )
        m1_ctl.value = pd.motor1.speed
        m1_ctl.motor = 1
        m1_ctl.bg = self._enabled_bg if pd.motor1.state else "lightgrey"
        m1_db = DebouncedSlider(self, pd.tmcc_id, 1, is_motor=True)
        m1_ctl.update_command(m1_db.on_change)
        widgets.append(m1_ctl)

        # make motor 2 on/off button
        m2_pwr, btn_h, btn_y = super()._make_state_button(pd, row, col + 1)
        m2_pwr.text = "Motor #2"
        m2_pwr.motor = 2
        if self.is_motor_active(pd, 2):
            self.set_button_active(m2_pwr)
        m2_pwr.update_command(self.set_motor_state, args=[pd.tmcc_id, 2])
        widgets.append(m2_pwr)

        # motor 2 control
        m2_ctl = Slider(
            self.btn_box,
            grid=[col + 1, row + 1],
            height=slider_height,
            width=self.pd_button_width,
            step=5,
        )
        m2_ctl.value = pd.motor2.speed
        m2_ctl.motor = 2
        m2_ctl.bg = self._enabled_bg if self.is_motor_active(pd, 2) else "lightgrey"
        m2_db = DebouncedSlider(self, pd.tmcc_id, 2, is_motor=True)
        m2_ctl.update_command(m2_db.on_change)
        widgets.append(m2_ctl)

        # make Lamp controls
        for lamp_no in range(1, 5):
            lamp = pd.get_lamp(lamp_no)
            if lamp_no % 2 == 1:
                row += 2
                lamp_col = col
            else:
                lamp_col = col + 1
            pwr, btn_h, btn_y = super()._make_state_button(pd, row, lamp_col)
            pwr.text = f"Lamp #{lamp_no}"
            pwr.lamp = lamp_no
            if lamp.level:
                self.set_button_active(pwr)
            pwr.update_command(self.set_lamp_state, args=[pd.tmcc_id, lamp_no])
            widgets.append(pwr)

            slider_height = int(round(btn_h * 0.9))
            ctl = Slider(
                self.btn_box,
                grid=[lamp_col, row + 1],
                height=slider_height,
                width=self.pd_button_width,
                step=5,
            )
            ctl.value = lamp.level
            ctl.lamp = lamp_no
            ctl.bg = self._enabled_bg if lamp.level else "lightgrey"
            dbs = DebouncedSlider(self, pd.tmcc_id, lamp_no, is_lamp=True)
            ctl.update_command(dbs.on_change)
            widgets.append(ctl)

        # noinspection PyTypeChecker
        self._state_buttons[pd.tmcc_id] = widgets
        return widgets, btn_h, btn_y


class DebouncedSlider:
    def __init__(
        self,
        gui: MotorsGui,
        tmcc_id: int,
        device: int,
        delay_ms=500,
        is_motor: bool = False,
        is_lamp: bool = False,
    ) -> None:
        self._is_lamp = is_lamp
        self._is_motor = is_motor
        self._gui = gui
        self._tmcc_id = tmcc_id
        self._device = device
        self._app = gui.app
        self._tk = gui.app.tk
        self._after_id = None
        self._delay = delay_ms
        self._last_value = None

    def on_change(self, value):
        if self._gui.is_making_buttons:
            return
        self._last_value = int(value)
        # Cancel previously scheduled call if any
        if self._after_id is not None:
            self._tk.after_cancel(self._after_id)
            self._after_id = None
        # Schedule new call after user stops moving for delay_ms
        self._after_id = self._tk.after(self._delay, self._fire)

    def _fire(self):
        self._after_id = None
        self.commit(self._last_value)

    def commit(self, value):
        # Do the real work once sliding has paused
        if self._is_motor:
            self._gui.set_motor_state(self._tmcc_id, self._device, value)
        elif self._is_lamp:
            self._gui.set_lamp_state(self._tmcc_id, self._device, value)



#
#  PyTrain: a library for controlling Lionel Legacy engines, trains, switches, and accessories
#
#  Copyright (c) 2024-2025 Dave Swindell <pytraininfo.gmail.com>
#
#  SPDX-License-Identifier: LPGL
#
#

from __future__ import annotations

from typing import cast

from ..db.component_state import SwitchState
from ..gui.component_state_gui import ComponentStateGui
from ..gui.state_based_gui import StateBasedGui
from ..protocol.command_req import CommandReq
from ..protocol.constants import CommandScope
from ..protocol.tmcc1.tmcc1_constants import TMCC1SwitchCommandEnum


class SwitchesGui(StateBasedGui):
    def __init__(
        self,
        label: str = None,
        width: int = None,
        height: int = None,
        aggrigator: ComponentStateGui = None,
        scale_by: float = 1.0,
    ) -> None:
        StateBasedGui.__init__(self, "Switches", label, width, height, aggrigator, disabled_bg="red", scale_by=scale_by)

    def get_target_states(self) -> list[SwitchState]:
        pds: list[SwitchState] = []
        accs = self._state_store.get_all(CommandScope.SWITCH)
        for acc in accs:
            acc = cast(SwitchState, acc)
            if acc.road_name and acc.road_name.lower() != "unused":
                pds.append(acc)
        return pds

    def is_active(self, state: SwitchState) -> bool:
        return state.is_thru

    def switch_state(self, pd: SwitchState) -> None:
        with self._cv:
            if pd.is_thru:
                CommandReq(TMCC1SwitchCommandEnum.OUT, pd.tmcc_id).send()
            else:
                CommandReq(TMCC1SwitchCommandEnum.THRU, pd.tmcc_id).send()



#
#  PyTrain: a library for controlling Lionel Legacy engines, trains, switches, and accessories
#
#  Copyright (c) 2024-2025 Dave Swindell <pytraininfo.gmail.com>
#
#  SPDX-License-Identifier: LPGL
#
#
from __future__ import annotations

import os
import shutil
import subprocess
import sys
from argparse import ArgumentParser
from pathlib import Path

from .make_base import _MakeBase
from .pytrain import DEFAULT_BUTTONS_FILE
from ..gui.component_state_gui import ComponentStateGui
from ..gui.accessories_gui import AccessoriesGui
from ..gui.launch_gui import LaunchGui
from ..gui.motors_gui import MotorsGui
from ..gui.power_district_gui import PowerDistrictsGui
from ..gui.routes_gui import RoutesGui
from ..gui.switches_gui import SwitchesGui
from ..gui.systems_gui import SystemsGui
from ..utils.argument_parser import PyTrainArgumentParser, UniqueChoice, IntRange
from ..utils.path_utils import find_file, find_dir

GUI_ARG_TO_CLASS = {
    "ac": AccessoriesGui,
    "accessories": AccessoriesGui,
    "co": ComponentStateGui,
    "component_state": ComponentStateGui,
    "la": LaunchGui,
    "launch_pad": LaunchGui,
    "mo": MotorsGui,
    "motors": MotorsGui,
    "pad": LaunchGui,
    "pd": PowerDistrictsGui,
    "po": PowerDistrictsGui,
    "power_districts": PowerDistrictsGui,
    "ro": RoutesGui,
    "routes": RoutesGui,
    "state": ComponentStateGui,
    "sw": SwitchesGui,
    "switches": SwitchesGui,
    "sy": SystemsGui,
    "systems": SystemsGui,
}

CLASS_TO_TEMPLATE = {
    AccessoriesGui: f"{AccessoriesGui.name()}(label=__LABEL__, scale_by=__SCALE_BY__)",
    ComponentStateGui: f"{ComponentStateGui.name()}(label=__LABEL__, initial=__INITIAL__, scale_by=__SCALE_BY__)",
    LaunchGui: f"{LaunchGui.__name__}(tmcc_id=__TMCC_ID__, track_id=__TRACK_ID__)",
    MotorsGui: f"{MotorsGui.name()}(label=__LABEL__, scale_by=__SCALE_BY__)",
    PowerDistrictsGui: f"{PowerDistrictsGui.name()}(label=__LABEL__, scale_by=__SCALE_BY__)",
    RoutesGui: f"{RoutesGui.name()}(label=__LABEL__, scale_by=__SCALE_BY__)",
    SwitchesGui: f"{SwitchesGui.name()}(label=__LABEL__, scale_by=__SCALE_BY__)",
    SystemsGui: f"{SystemsGui.name()}(label=__LABEL__, scale_by=__SCALE_BY__, hold_for=__HOLD_FOR__)",
}

NEED_FONTS = {
    LaunchGui,
}

CHOICES = [
    "accessories",
    "motors",
    "power districts",
    "routes",
    "switches",
    "systems",
]

CHOICES_HELP = ", ".join([x.title() for x in CHOICES])


class MakeGui(_MakeBase):
    """
    Handles the creation, configuration, and management of a graphical user interface (GUI) for PyTrain.

    This class facilitates the configuration and installation of various GUI components as specified by command-line
    arguments. It is designed to be extensible to support multiple GUI types, offering flexibility for user interaction
    and program operations. The functionality includes argument preprocessing, GUI setup, installation, and removal,
    as well as command-line parser configuration tailored to specific GUI requirements.
    """

    def __init__(self, cmd_line: list[str] = None) -> None:
        self._start_gui = False
        self._launch_path = self._desktop_path = self._buttons_path = self._fonts_path = None
        self._imports = self._gui_class = self._gui_stmt = None
        self._gui_config = dict()
        super().__init__(cmd_line)

    def program(self) -> str:
        return "make_gui"

    def function(self) -> str:
        return "GUI"

    def postprocess_args(self) -> None:
        from .. import PROGRAM_BASE, is_package

        if not self._args.remove and not self._args.gui:
            self._parser.error("the following arguments are required: gui")
        if self._args.start is True:
            self._start_gui = True
        self._buttons_file = DEFAULT_BUTTONS_FILE
        self._launch_path = Path(self._home, "launch_pytrain.bash")
        self._desktop_path = Path(self._home, ".config", "autostart", "pytrain.desktop")
        self._buttons_path = Path(self._cwd, self._buttons_file)
        self._fonts_path = Path(self._home, ".fonts")
        self._imports = f"from {PROGRAM_BASE if is_package() else 'src.' + PROGRAM_BASE} import *"
        self._gui_class = GUI_ARG_TO_CLASS.get(self._args.gui)
        self.harvest_gui_config()

    def postprocess_config(self) -> None:
        self._config["___IMPORTS___"] = self._imports
        self._config["___GUI___"] = self._gui_stmt = self.construct_gui_stmt()

    def config_header(self) -> list[str]:
        from .. import PROGRAM_NAME

        lines = list()
        lines.append(f"\nInstalling the {PROGRAM_NAME} {self._gui_class.__name__} GUI with these settings:")
        lines.append(f"  Start GUI now: {'Yes' if self._start_gui is True else 'No'}")
        lines.append(f"  Imports: {self._imports}")
        lines.append(f"  GUI: {self._gui_stmt}")
        return lines

    def install(self) -> None:
        from .. import PROGRAM_NAME

        path = self.make_shell_script()
        if path:
            self._config["___SHELL_SCRIPT___"] = str(path)
        else:
            return

        desktop = self.make_python_desktop_file()
        if desktop:
            self._config["___DESKTOP___"] = str(desktop)
        else:
            return

        buttons = self.make_buttons_file()
        if buttons:
            self._config["___BUTTONS___"] = str(buttons)
        else:
            return

        if self._gui_class in NEED_FONTS:
            self.install_fonts()

        if self._start_gui:
            self.spawn_detached(path)
            print(f"\nStarting {PROGRAM_NAME} {self._gui_class.__name__} GUI...")

    def remove(self) -> None:
        from .. import PROGRAM_NAME

        if not self.is_gui_present:
            print(f"\nNo {PROGRAM_NAME} GUI detected. Exiting")
            return
        if self.confirm(f"Are you sure you want to remove the {PROGRAM_NAME} GUI?"):
            self.find_and_kill_process(cmdline="python3 headless -buttons")
            for path in (self._desktop_path, self._launch_path, self._buttons_path):
                if path.exists():
                    print(f"\nRemoving {path}...")
                    path.unlink(missing_ok=True)

    # noinspection PyTypeChecker
    def command_line_parser(self) -> ArgumentParser:
        from .. import PROGRAM_NAME

        parser = ArgumentParser(add_help=False)
        sp = parser.add_subparsers(dest="gui", help="Available GUIs")

        # Launch Pad GUI
        pad = sp.add_parser(
            "launch_pad",
            aliases=["la", "pad"],
            allow_abbrev=True,
            help="Launch Pad GUI",
        )
        pad.add_argument(
            "-tmcc_id",
            type=IntRange(1, 98),
            default=39,
            const=39,
            nargs="?",
            help="Launch Pad TMCC ID (default: 39)",
        )
        pad.add_argument(
            "-track_id",
            type=IntRange(1, 98),
            help="Launch Pad Track Power District TMCC ID",
        )

        # Component State GUI
        comp = sp.add_parser(
            "component_state",
            aliases=["co", "state"],
            allow_abbrev=True,
            help="Component State GUI",
        )

        comp.add_argument(
            "-initial",
            type=UniqueChoice(CHOICES),
            nargs="?",
            const="power districts",
            default="power districts",
            help=f"Initial Display (default: Power Districts, choices: {CHOICES_HELP})",
        )
        comp.add_argument(
            "-label",
            type=str,
            help="Layout Name",
        )
        comp.add_argument(
            "-scale_by",
            type=float,
            default=1.0,
            help="Text Scale Factor (default: 1.0)",
        )

        # Accessories GUI
        acc = sp.add_parser(
            "accessories",
            aliases=["ac"],
            allow_abbrev=True,
            help="Accessories GUI",
        )

        acc.add_argument(
            "-label",
            type=str,
            help="Layout Name",
        )
        acc.add_argument(
            "-scale_by",
            type=float,
            default=1.0,
            help="Text Scale Factor (default: 1.0)",
        )

        # Motors GUI
        mo = sp.add_parser(
            "motors",
            aliases=["mo"],
            allow_abbrev=True,
            help="Motors GUI",
        )
        mo.add_argument(
            "-label",
            type=str,
            help="Layout Name",
        )
        mo.add_argument(
            "-scale_by",
            type=float,
            default=1.0,
            help="Text Scale Factor (default: 1.0)",
        )

        # Power Districts GUI
        pd = sp.add_parser(
            "power_districts",
            aliases=["po", "pd"],
            allow_abbrev=True,
            help="Power Districts GUI",
        )
        pd.add_argument(
            "-label",
            type=str,
            help="Layout Name",
        )
        pd.add_argument(
            "-scale_by",
            type=float,
            default=1.0,
            help="Text Scale Factor (default: 1.0)",
        )

        # Routes GUI
        ro = sp.add_parser(
            "routes",
            aliases=["ro"],
            allow_abbrev=True,
            help="Routes GUI",
        )
        ro.add_argument(
            "-label",
            type=str,
            help="Layout Name",
        )
        ro.add_argument(
            "-scale_by",
            type=float,
            default=1.0,
            help="Text Scale Factor (default: 1.0)",
        )

        # Switches GUI
        sw = sp.add_parser(
            "switches",
            aliases=["sw"],
            allow_abbrev=True,
            help="Switches GUI",
        )
        sw.add_argument(
            "-label",
            type=str,
            help="Layout Name",
        )
        sw.add_argument(
            "-scale_by",
            type=float,
            default=1.0,
            help="Text Scale Factor (default: 1.0)",
        )

        # Systems GUI
        sy = sp.add_parser(
            "systems",
            aliases=["sy"],
            allow_abbrev=True,
            help=f"{PROGRAM_NAME} Systems Administration GUI",
        )
        sy.add_argument(
            "-label",
            type=str,
            help="Layout Name",
        )
        sy.add_argument(
            "-scale_by",
            type=float,
            default=1.0,
            help="Text Scale Factor (default: 1.0)",
        )
        sy.add_argument(
            "-hold_for",
            type=IntRange(1, 10),
            default=5,
            const=5,
            nargs="?",
            help="Hold time before performing action (default: 5 seconds)",
        )

        misc_opts = parser.add_argument_group("Service options")
        misc_opts.add_argument(
            "-start",
            action="store_true",
            help=f"Start {PROGRAM_NAME} GUI now (otherwise, it starts on reboot)",
        )
        return PyTrainArgumentParser(
            prog=self._prog,
            description=f"Launch {PROGRAM_NAME} GUI when your Raspberry Pi is powered on",
            parents=[self._command_line_parser(), parser],
        )

    def make_shell_script(self) -> Path | None:
        template = find_file("launch_pytrain.bash.template", (".", "../", "src"))
        if template is None:
            print("\nUnable to locate shell script template. Exiting")
            return None
        template_data = ""
        with open(template, "r") as f:
            template_data = f.read()
        for key, value in self.config.items():
            template_data = template_data.replace(key, value)
        path = self._launch_path
        # write the shell script file
        if path.exists():
            shutil.copy2(path, path.with_suffix(".bak"))
        with open(path, "w") as f:
            f.write(template_data)
        os.chmod(path, 0o755)
        print(f"\n{path} created")
        return path

    def make_python_desktop_file(self) -> Path | None:
        template = find_file("pytrain_desktop.template", (".", "../", "src"))
        if template is None:
            print("\nUnable to locate desktop template. Exiting")
            return None
        template_data = ""
        with open(template, "r") as f:
            template_data = f.read()
        for key, value in self.config.items():
            template_data = template_data.replace(key, value)
        # make sure directory exists
        path = self._desktop_path
        if not path.parent.exists():
            path.parent.mkdir(parents=True, exist_ok=True)
            os.chmod(path.parent, 0o755)
        # write the desktop file
        with open(path, "w") as f:
            f.write(template_data)

        print(f"\n{path} created")
        return path

    def make_buttons_file(self) -> Path | None:
        template = find_file("buttons_gui.py.template", (".", "../", "src"))
        if template is None:
            print("\nUnable to locate buttons template. Exiting")
            return None
        template_data = ""
        with open(template, "r") as f:
            template_data = f.read()
        for key, value in self.config.items():
            template_data = template_data.replace(key, value)
        # make sure directory exists
        path = self._buttons_path
        # write the buttons file
        if path.exists():
            shutil.copy2(path, path.with_suffix(".bak"))
        with open(path, "w") as f:
            f.write(template_data)

        print(f"\n{path} created")
        return path

    def install_fonts(self) -> Path | None:
        template = find_dir("fonts", (".", "../", "src"))
        if template is None:
            print("\nUnable to locate fonts directory. Exiting")
            return None
        path = self._fonts_path
        if not path.exists():
            path.mkdir(parents=True, exist_ok=True)
        # copy the fonts directory
        try:
            shutil.copytree(template, path, dirs_exist_ok=True)
            subprocess.run(["fc-cache", "-f", "-v"])
            print(f"Installed fonts to: {path}")
        except shutil.Error as e:
            print(f"Error copying directory: {e}")
            path = None
        except OSError as e:
            print(f"Error: {e}")
            path = None
        return path

    def harvest_gui_config(self):
        if hasattr(self._args, "initial"):
            self._gui_config["__INITIAL__"] = f"'{self._args.initial.title()}'" if self._args.initial else "None"
        if hasattr(self._args, "label"):
            self._gui_config["__LABEL__"] = f"'{self._args.label}'" if self._args.label else "None"
        if hasattr(self._args, "scale_by"):
            self._gui_config["__SCALE_BY__"] = str(self._args.scale_by)
        if hasattr(self._args, "hold_for"):
            self._gui_config["__HOLD_FOR__"] = str(self._args.hold_for)
        if hasattr(self._args, "tmcc_id"):
            self._gui_config["__TMCC_ID__"] = str(self._args.tmcc_id)
        if hasattr(self._args, "track_id"):
            self._gui_config["__TRACK_ID__"] = str(self._args.track_id)

    def construct_gui_stmt(self):
        stmt = CLASS_TO_TEMPLATE.get(self._gui_class)
        for key, value in self._gui_config.items():
            stmt = stmt.replace(key, value)
        return stmt

    @property
    def is_gui_present(self) -> bool:
        launch_path = Path(self._home, "launch_pytrain.bash")
        desktop_path = Path(self._home, ".config", "autostart", "pytrain.desktop")
        return launch_path.exists() and desktop_path.exists()


def main(args: list[str] | None = None) -> int:
    if args is None:
        args = sys.argv[1:]
    try:
        MakeGui(args)
        return 0
    except Exception as e:
        # Output anything else nicely formatted on stderr and exit code 1
        sys.exit(f"{__file__}: error: {e}\n")



#
#  PyTrain: a library for controlling Lionel Legacy engines, trains, switches, and accessories
#
#  Copyright (c) 2024-2025 Dave Swindell <pytraininfo.gmail.com>
#
#  SPDX-License-Identifier: LPGL
#
#

from __future__ import annotations

from threading import Event
from typing import cast

from guizero.event import EventData

from ..protocol.tmcc1.tmcc1_constants import TMCC1AuxCommandEnum
from ..protocol.command_req import CommandReq
from ..protocol.constants import CommandScope
from ..db.accessory_state import AccessoryState
from ..gui.component_state_gui import ComponentStateGui
from ..gui.state_based_gui import StateBasedGui, MomentaryActionHandler
from ..pdi.asc2_req import Asc2Req
from ..pdi.constants import PdiCommand, Asc2Action


class AccessoriesGui(StateBasedGui):
    def __init__(
        self,
        label: str = None,
        width: int = None,
        height: int = None,
        aggrigator: ComponentStateGui = None,
        scale_by: float = 1.0,
    ) -> None:
        self._is_momentary = set()
        self._released_events = dict[int, Event]()
        StateBasedGui.__init__(self, "Accessories", label, width, height, aggrigator, scale_by=scale_by)

    def _post_process_state_buttons(self) -> None:
        for tmcc_id in self._is_momentary:
            if tmcc_id in self._state_buttons:
                pb = self._state_buttons[tmcc_id]
                pb.when_left_button_pressed = self.when_pressed
                pb.when_left_button_released = self.when_released

    def get_target_states(self) -> list[AccessoryState]:
        pds: list[AccessoryState] = []
        accs = self._state_store.get_all(CommandScope.ACC)
        for acc in accs:
            acc = cast(AccessoryState, acc)
            if acc.is_power_district or acc.is_sensor_track or acc.is_amc2:
                continue
            if acc.road_name and acc.road_name.lower().strip() != "unused":
                pds.append(acc)
                name_lc = acc.road_name.lower()
                if "aux1" in name_lc or "ax1" in name_lc or "(a1)" in name_lc or "(m)" in name_lc:
                    self._is_momentary.add(acc.address)
        return pds

    def is_active(self, state: AccessoryState) -> bool:
        return state.is_aux_on

    def switch_state(self, pd: AccessoryState) -> None:
        with self._cv:
            if pd.tmcc_id in self._is_momentary:
                pass
            elif pd.is_aux_on:
                CommandReq(TMCC1AuxCommandEnum.AUX2_OPT_ONE, pd.tmcc_id).send()
            else:
                CommandReq(TMCC1AuxCommandEnum.AUX2_OPT_ONE, pd.tmcc_id).send()

    def when_pressed(self, event: EventData) -> None:
        pb = event.widget
        state = pb.component_state
        if state.is_asc2:
            Asc2Req(state.address, PdiCommand.ASC2_SET, Asc2Action.CONTROL1, values=1).send()
        else:
            if state.tmcc_id in self._released_events:
                event = self._released_events[state.tmcc_id]
                self._released_events[state.tmcc_id].clear()
            else:
                self._released_events[state.tmcc_id] = event = Event()
            _ = MomentaryActionHandler(pb, event, state, 0.2)

    def when_released(self, event: EventData) -> None:
        state = event.widget.component_state
        if state.is_asc2:
            Asc2Req(state.address, PdiCommand.ASC2_SET, Asc2Action.CONTROL1, values=0).send()
        else:
            state = event.widget.component_state
            self._released_events[state.tmcc_id].set()
